# A function defined in this s-file:
#    quad3 : integrator of f(x) over (a,infinity), 
#                 f(x) is non oscillatory function.
#
#    (C) Shin'ichi OISHI  ---2001/8/25---
#
#    DE-Quadrature
#    Numerical Automatic Integrator for Improper Integral
#    method    : Double Exponential (DE) Transformation
#    dimension : one
#    [description]
#        I = integral of f(x) over (a,infinity), 
#            f(x) has not oscillatory factor.
#    [usage]
#        quad3(a, eps)
#    [parameters]
#        a         : lower limit of integration (double)
#        eps       : relative error requested (double)
#    [answer]
#        i         : approximation to the integral (double *)
#        err       : estimate of the absolute error (double *)
#    [remarks]
#        function
#            f(x) needs to be analytic over (a,infinity).
#        relative error
#            eps is relative error requested excluding 
#            cancellation of significant digits.
#            i.e. eps means : (absolute error) / 
#                             (integral_a^infinity |f(x)| dx).
#            eps does not mean : (absolute error) / I.
#        error message
#            err >= 0 : normal termination.
#            err < 0  : abnormal termination (m >= mmax).
#                       i.e. convergent error is detected :
#                           1. f(x) or (d/dx)^n f(x) has 
#                              discontinuous points or sharp 
#                              peaks over (a,infinity).
#                              you must divide the interval 
#                              (a,infinity) at this points.
#                           2. relative error of f(x) is 
#                              greater than eps.
#                           3. f(x) has oscillatory factor 
#                              and decay of f(x) is very slow 
#                              as x -> infinity.

function quad3(f, a, eps) {
    mmax = 256;
    efs = 0.1;
    hoff = 11.0;
    pi4 = atan(1.0);
    epsln = 1 - log(efs * eps);
    epsh = sqrt(efs * eps);
    h0 = hoff / epsln;
    ehp = exp(h0);
    ehm = 1 / ehp;
    epst = exp(-ehm * epsln);
    ir = eval(f|a + 1);
    i = ir * (2 * pi4);
    err = fabs(i) * epst;
    h = 2 * h0;
    m = 1;
    do {
        iback = i;
        irback = ir;
        t = h * 0.5;
        do {
            em = exp(t);
            ep = pi4 * em;
            em = pi4 / em;
            do {
                xp = exp(ep - em);
                xm = 1 / xp;
                fp = eval(f|a + xp) * xp;
                fm = eval(f|a + xm) * xm;
                ir = ir + fp + fm;
                i  = i + (fp + fm) * (ep + em);
                errt = (fabs(fp) + fabs(fm)) * (ep + em);
                if (m == 1) {err = err + errt * epst};
                ep = ep * ehp;
                em = em * ehm;
            } while ((errt > err) || (xm > epsh));
            t = t + h;
        } while (t < h0);
        if (m == 1) {
            errh = (err / epst) * epsh * h0;
            errd = 1 + 2 * errh;
        } else {
            errd = h * (fabs(i - 2 * iback) + 4 * fabs(ir - 2 * irback));
        };
        h = h * 0.5;
        m = m * 2;
    } while ((errd > errh) && (m < mmax));
    i = i * h;
    if (errd > errh) {
        err = -errd * m;
    } else {
        err = errh * epsh * m / (2 * efs);
    };
    <i,err>;
}
