# A function defined in this s-file:
#    quad4 : integrator of f(x) over (a,infinity), 
#                 f(x) is oscillatory function.
#
#    (C) Shin'ichi OISHI   ---2001/8/25---
#    DE-Quadrature
#    Numerical Automatic Integrator for Improper Integral
#    method    : Double Exponential (DE) Transformation
#    dimension : one
#    [description]
#        I = integral of f(x) over (a,infinity), 
#            f(x) has oscillatory factor :
#            f(x) = g(x) * sin(omega * x + theta) as x -> infinity.
#    [usage]
#        quad4(a, omega, eps);
#    [parameters]
#        a         : lower limit of integration (double)
#        omega     : frequency of oscillation (double)
#        eps       : relative error requested (double)
#    [ansewer]
#        i         : approximation to the integral (double *)
#        err       : estimate of the absolute error (double *)
#    [remarks]
#        function
#            f(x) needs to be analytic over (a,infinity).
#        relative error
#            eps is relative error requested excluding 
#            cancellation of significant digits.
#            i.e. eps means : (absolute error) / 
#                             (integral_a^R |f(x)| dx).
#            eps does not mean : (absolute error) / I.
#        error message
#            err >= 0 : normal termination.
#            err < 0  : abnormal termination (m >= mmax).
#                       i.e. convergent error is detected :
#                           1. f(x) or (d/dx)^n f(x) has 
#                              discontinuous points or sharp 
#                              peaks over (a,infinity).
#                              you must divide the interval 
#                              (a,infinity) at this points.
#                           2. relative error of f(x) is 
#                              greater than eps.

function quad4(f, a, omega, eps) {
    mmax = 256;
    lmax = 5;
    efs = 0.1;
    enoff = 0.40;
    pqoff = 2.9;
    ppoff = -0.72;
    pi4 = atan(1.0);
    epsln = 1 - log(efs * eps);
    epsh = sqrt(efs * eps);
    n = enoff * epsln;
    frq4 = fabs(omega) / (2 * pi4);
    per2 = 4 * pi4 / fabs(omega);
    pq = pqoff / epsln;
    pp = ppoff - log(pq * pq * frq4);
    ehp = exp(2 * pq);
    ehm = 1 / ehp;
    xw = exp(pp - 2 * pi4);
    i = eval(f|a + sqrt(xw * (per2 * 0.5)));
    ir = i * xw;
    i = i * (per2 * 0.5);
    err = fabs(i);
    h = 2;
    m = 1;
    do {
        iback = i;
        irback = ir;
        t = h * 0.5;
        do {
            em = exp(2 * pq * t);
            ep = pi4 * em;
            em = pi4 / em;
            tk = t;
            do {
                xw = exp(pp - ep - em);
                wg = sqrt(frq4 * xw + tk * tk);
                xa = xw / (tk + wg);
                wg = (pq * xw * (ep - em) + xa) / wg;
                fm = eval(f|a + xa);
                fp = eval(f|a + xa + per2 * tk);
                ir = ir + (fp + fm) * xw;
                fm = fm * wg;
                fp = fp * (per2 - wg);
                i = i + fp + fm;
                if (m == 1) {err = err + fabs(fp) + fabs(fm)};
                ep = ep * ehp;
                em = em * ehm;
                tk = tk + 1;
            } while (ep < epsln);
            if (m == 1) {
                errh = err * epsh;
                err = err * eps;
            };
            tn = tk;
            while (fabs(fm) > err) {
                xw = exp(pp - ep - em);
                xa = xw / tk * 0.5;
                wg = xa * (1 / tk + 2 * pq * (ep - em));
                fm = eval(f|a + xa);
                ir = ir + fm * xw;
                fm = fm * wg;
                i =  i + fm;
                ep = ep * ehp;
                em = em * ehm;
                tk = tk + 1;
            };
            fm = eval(f|a + per2 * tn);
            em = per2 * fm;
            i = i + em;
            if ((fabs(fp) > err) || (fabs(em) > err)) {
                l = 0;
                do {
                    l = l + 1;
                    tn = tn + n;
                    em = fm;
                    fm = eval(f|a + per2 * tn);
                    xa = fm;
                    ep = fm;
                    em = em + fm;
                    xw = 1;
                    wg = 1;
                    for (k = 1; k <= n - 1; k = k+1) {
                        xw = xw * (n + 1 - k) / k;
                        wg = wg + xw;
                        fp = eval(f|a + per2 * (tn - k));
                        xa = xa + fp;
                        ep = ep + fp * wg;
                        em = em + fp * xw;
                    };
                    wg = per2 * n / (wg * n + xw);
                    em = wg * fabs(em);
                    i = i + per2 * xa;
                } while((em > err) && (l < lmax));
                i = i - per2 * xa;
                i = i + wg * ep;
                if (em > err) {err = em};
            };
            t = t + h;
        } while (t < 1);
        if (m == 1) {
            errd = 1 + 2 * errh;
        } else {
            errd = h * (fabs(i - 2 * iback) + pq * fabs(ir - 2 * irback));
        };
        h = h * 0.5;
        m = m * 2;
    } while ((errd > errh) && (m < mmax));
    i = i * h;
    if (errd > errh) {
        err = -errd;
    } else {
        err = err * m * 0.5;
    };
    <i, err>;
}
