#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "SLAB_MEMORY.h"
#define GLOBAL_VARIABLE_DEFINE
#include "slab.h"

void*
SLAB_MEMORY_malloc(size_t size)
{
    void	*ptr;

    if ((ptr = GC_malloc(size)) == NULL) {
        fprintf(stderr, "memory error\n");
    }

    return(ptr);
}

void*
SLAB_MEMORY_realloc(void *ptr,size_t size)
{
    if (ptr == NULL) {
	if ((ptr = GC_malloc(size)) == NULL) {
            fprintf(stderr, "memory error\n");
	}
    } else {
	if ((ptr = GC_realloc(ptr,size)) == NULL) {
            fprintf(stderr, "memory error\n");
	}
    }
    return(ptr);
}

void
SLAB_MEMORY_free(void *ptr)
{
    if (ptr != NULL) {
	GC_free(ptr);
    }
}

typedef union {
    long	l_dummy;
    double	d_dummy;
    void	*p_dummy;
} Cell;

#define	CELL_SIZE		(sizeof(Cell))
#define DEFAULT_PAGE_SIZE	(4096)	/* cell num */

typedef struct MemoryPage_tag MemoryPage;
typedef MemoryPage *MemoryPageList;

struct MemoryPage_tag {
    int			cell_num;
    int			use_cell_num;
    MemoryPageList	next;
    Cell		cell[1];
};

struct MEM_Storage_tag {
    MemoryPageList	page_list;
    int			current_page_size;
};

MEM_Storage
MEM_open_storage(int page_size)
{
    MEM_Storage	storage;

    storage = SLAB_MEMORY_malloc(sizeof(struct MEM_Storage_tag));
    storage->page_list = NULL;
    assert(page_size >= 0);
    if (page_size > 0) {
	storage->current_page_size = page_size;
    } else {
	storage->current_page_size = DEFAULT_PAGE_SIZE;
    }

    return storage;
}

void*
MEM_storage_malloc(MEM_Storage storage, size_t size)
{
    int			cell_num;
    MemoryPage		*new_page;
    void 		*p;

    cell_num = ((size - 1) / CELL_SIZE) + 1;

    if (storage->page_list != NULL
	&& (storage->page_list->use_cell_num + cell_num
	    < storage->page_list->cell_num)) {
	p = &(storage->page_list->cell[storage->page_list->use_cell_num]);
	storage->page_list->use_cell_num += cell_num;
    } else {
	int	alloc_cell_num;

	alloc_cell_num = larger(cell_num, storage->current_page_size);

	new_page = SLAB_MEMORY_malloc(sizeof(MemoryPage)
				   + CELL_SIZE * (alloc_cell_num - 1));
	new_page->next = storage->page_list;
	new_page->cell_num = alloc_cell_num;
	storage->page_list = new_page;

	p = &(new_page->cell[0]);
	new_page->use_cell_num = cell_num;
    }

    return p;
}

void
MEM_dispose_storage(MEM_Storage storage)
{
    MemoryPage	*temp;

    while (storage->page_list) {
	temp = storage->page_list->next;

	SLAB_MEMORY_free(storage->page_list);
	storage->page_list = temp;
    }
}

void *
clc_malloc(size_t size)
{
    void *p;

    p = MEM_storage_malloc(clc_current_interpreter->current_storage, size);

    return p;
}

void *
clc_global_malloc(size_t size)
{
    void *p;

    p = MEM_storage_malloc(clc_current_interpreter->global_storage, size);

    return p;
}

Value *
clc_search_local_variable(LocalEnvironment *env, char *identifier)
{
    Variable	*pos;

    if (env == NULL)
	return NULL;
    for (pos = env->variable; pos; pos = pos->next) {
	if (!strcmp(pos->name, identifier))
	    break;
    }
    if (pos == NULL) {
	return NULL;
    } else {
	return &pos->value;
    }
}

Value *
clc_search_global_variable(char *identifier)
{
    Variable	*pos;

    for (pos = clc_current_interpreter->variable;
	 pos; pos = pos->next) {
	if (!strcmp(pos->name, identifier))
	    break;
    }
    if (pos == NULL) {
	return NULL;
    } else {
	return &pos->value;
    }
}

void
clc_add_local_variable(LocalEnvironment *env,
		       char *identifier, Value *value)
{
    Variable	*new_variable;

    new_variable = SLAB_MEMORY_malloc(sizeof(Variable));
    new_variable->name = identifier;
    new_variable->value = *value;
    new_variable->next = env->variable;
    env->variable = new_variable;
}

void
clc_add_global_variable(char *identifier, Value *value)
{
    Variable	*new_variable;
    new_variable = clc_global_malloc(sizeof(Variable));
    new_variable->name = clc_global_malloc(strlen(identifier) + 1);
    strcpy(new_variable->name, identifier);
    new_variable->value = *value;

    new_variable->next = clc_current_interpreter->variable;
    clc_current_interpreter->variable = new_variable;
}

FunctionDefinition *
clc_search_function(char *name)
{
    FunctionDefinition *pos;
    for (pos = clc_current_interpreter->function_list;
	 pos; pos = pos->next) {
	if (!strcmp(pos->name, name))
	    break;
    }
    return pos;
}

void
clc_reopen_current_storage(void)
{
    MEM_dispose_storage(clc_current_interpreter->current_storage);
    clc_current_interpreter->current_storage = MEM_open_storage(0);
}



SLAB_Interpreter
SLAB_create_interpreter(void)
{
    MEM_Storage	storage = MEM_open_storage(0);

    SLAB_Interpreter interpreter
	= MEM_storage_malloc(storage, sizeof(struct SLAB_Interpreter_tag));
    interpreter->input_mode = SLAB_TTY_INPUT_MODE;
    interpreter->input_fp = NULL;
    interpreter->current_storage = NULL;
    interpreter->global_storage = storage;
    interpreter->function_list = NULL;

    return interpreter;
}

void
SLAB_interpret(SLAB_Interpreter interpreter, FILE *fp, SLAB_InputMode input_mode)
{
    extern int yyparse(void);

    interpreter->input_mode = input_mode;
    interpreter->input_fp = fp;

    interpreter->current_storage = MEM_open_storage(0);
    clc_current_interpreter = interpreter;

    if (yyparse()) {
	fprintf(stderr, "Error ! Error ! Error !\n");
	exit(1);
    }

    if (interpreter->current_storage) {
	MEM_dispose_storage(interpreter->current_storage);
    }
}

void
SLAB_dispose_interpreter(SLAB_Interpreter interpreter)
{
    MEM_dispose_storage(interpreter->current_storage);

    while (interpreter->function_list) {
	FunctionDefinition	*temp;
	temp = interpreter->function_list->next;
	MEM_dispose_storage(interpreter->function_list->storage);
	interpreter->function_list = temp;
    }
}
